* Disclaimer:
* This source code is proprietary of 
* Integrated Signal Processing Systems / RWTH-Aachen / Prof. Meyr (ISS)
* It is part of the DSPstone Benchmarking methodology
* The code, or parts of it may be used for non commercial issues
* All commercial uses require the written consent of  ISS

* file:			lms.asm
* author :		Heiner Loellmann / Andreas Ropers
* date :		22.06.1999
* processor :		TMS320C54x
* project :		dsp-stone
* short description:	realisation of an adaptive filter with the help
*			of the LMS Algorithm
*
*  
*               x(n-0)  ---- x(n-1)  ----- x(n-i) -----  x(n-N+1)
*    x(n) --->---------| 1/z|-------| 1/z|-.....-| 1/z|----
*                  |   -----    |   ------   |   ------   |
*                  h0           h1           hi          hN-1
*                    \________  |____   _____|  _________/
*                             \______(+)_______/
*                                     |
*                                     |---------> y(n)
*                                     v
*          d(n) -------------------->(+)--------> e(n)
*                                   -
* 
* 
* Notation and symbols:
* x(n)  - Input sample at time n.
* d(n)  - Desired signal at time n.
* y(n)  - FIR filter output at time n.
* H(n)  - Filter coefficient vector at time n.  H={h0,h1,..,hN-1}
* X(n)  - Filter state variable vector at time N.
*         X={x(0),x(n-1),...,x(n-N+1)}
* delta - Adaptation gain.
* N     - Number of coefficient taps in the filter. 
* 
* PROCESSING:
*
*     True LMS Algorithm       
*     ------------------       
*     Get input sample         
*     Save input sample        
*     Do FIR                   
*     Shift vector X           
*     Get d(n), calculate e(n)      
*     Update coefficients      
*     Output y(n)              
*
*
* System equations:
* 
* e(n)  = d(n) - H(n)  X(n)       (FIR filter and error)
* H(n+1)= H(n) + delta X(n) e(n)  (Coefficient update)
* 
* References:
*  "Adaptive Digital Filters and Signal Analysis", Maurice G. Bellanger
*     Marcel Dekker, Inc. New York and Basel
*  
*   "The DLMS Algorithm Suitable for the Pipelined Realization of Adaptive
*      Filters", Proc. IEEE ASSP Workshop, Academia Sinica, Beijing, 1986
*

	.global startp
	.global prof
	.global endp
	.mmregs

N	.set 4h			; number of filter-coefficients

	.data

h	.word 1h		; h(0)
	.word 2h
	.word 3h
	.word 4h		; h(N-1)
	
x	.word 1h		; x(n)
	.word 2h
	.word 3h		; x(n-(N-2))
	.word 4h		; x(n-(N-1))
	
y	.word 0h		; y(n)
	
d	.word 14h		; d(n)  (desired signal)
	
delta	.word 2h		; Adaptation gain
	
error	.word 0h		; error(n)=y(n)-d(n)
	
new_x	.word 1h		; new input value x(n)

	
	.text
*	assumption:	FRCT = 0
startp:	
	STM #y     , AR1	; pointer to y(n)
	STM #x+2   , AR2	; pointer to x(n-(N-2))
	STM #x+3   , AR3	; pointer to x(n-(N-1))
	STM #h+3   , AR4	; pointer to h(N-1)
prof:	
			; y(n) = sum (i=0...N-1) of (x(n-1)*h(i))
	STM #N-2   , BRC	
	LD  #0     , B
	RPTB conv		; for i= 0...N-2 do {
	LMS *AR4-  , *AR2		; B = B + h(1-N+i)*x(n-(N-2)+i)
conv	MVDD *AR2- , *AR3-	; x(n-(N-2)+i) -> x(n-(N-1)+i)  }

	STM #new_x , AR5
	MVDD *AR5  , *AR3	; get the new input sample x(n)
	LMS *AR3   , *AR4	; the last iteration B = B+x(n)*h(0)
	STL B      , *AR1	; B -> y(n)

			; error = (d(n)-y(n))*delta
			; (weighted difference between desired and calculated signal )
	STM #error , AR5
	STM #delta , AR6
	STM #d     , AR7
	
	SUB *AR7   , B		; B = y(n) - d(n)
	NEG B			; B = d(n) - y(n)
	STL B      , *AR5	; B -> error
	LD  *AR5   , T		; error -> TREG    ( easier way possible ?)
	MPY *AR6   , B		; B = delta*error
	STL B      , *AR5	; B -> error

			; h(i) = h(i)+error*x(n-i) whith i=0...N-1 
	STM #N-1   , BRC
	RPTB update		; for i=0...N-1 do {
	LD  *AR4   , B		; h(i) -> B
	LMS *AR3+  , *AR5	; B = B + error*x(n-i)
update	STL  B     , *AR4+	; h(i) = B  }
endp:	nop
	
	.end
